local MAJOR, MINOR = "LibInventory-2.1", 5000401 -- 5.0.4 v1
local LibInventory = LibStub:NewLibrary(MAJOR, MINOR)

if not LibInventory then return end -- no need to update

local CallbackHandler = LibStub:GetLibrary("CallbackHandler-1.0")

-- to populate the cache with the auction house, run this when the AH is opened:
-- /script QueryAuctionItems("", nil, nil, 0, 0, 0, 0, 0, 0, true)

-- /script local maxid = 0; for i = 91000, 92000 do if GetItemIcon(i) then maxid = i end end; print(maxid)

-- -----
-- local upvalues
-- -----
local _G = getfenv(0)

local collectgarbage		= _G.collectgarbage
local ipairs				= _G.ipairs
local min					= _G.min
local next					= _G.next
local pairs					= _G.pairs
local select				= _G.select
local tonumber				= _G.tonumber
local tostring				= _G.tostring
local type					= _G.type

local GetItemIcon			= _G.GetItemIcon
local GetItemInfo			= _G.GetItemInfo
local InCombatLockdown		= _G.InCombatLockdown

local strfind				= _G.string.find
local strformat				= _G.string.format
local strgmatch				= _G.string.gmatch
local strmatch				= _G.string.match

local table_concat			= _G.table.concat
local table_insert			= _G.table.insert
local table_remove			= _G.table.remove
local table_sort			= _G.table.sort
local table_wipe			= _G.table.wipe


-- -----
-- lib variables
-- -----
LibInventory.frame					= LibInventory.frame				or CreateFrame("Frame", "LibInventory21Frame") -- our event frame
LibInventory.callbacks				= LibInventory.callbacks			or CallbackHandler:New(LibInventory)

LibInventory.items					= LibInventory.items				or {}
LibInventory.items_rescan			= LibInventory.items_rescan			or {}
LibInventory.items_valid_ids		= LibInventory.items_valid_ids		or {}
LibInventory.item_last_scanned		= LibInventory.item_last_scanned	or nil
LibInventory.count_known			= LibInventory.count_known			or 0
LibInventory.count_total			= LibInventory.count_total			or 0
LibInventory.highest_id				= LibInventory.highest_id			or 0
LibInventory.types_base				= LibInventory.types_base			or {}
LibInventory.types_sub				= LibInventory.types_sub			or {}
LibInventory.types_slot				= LibInventory.types_slot			or {}
LibInventory.types_tree				= LibInventory.types_tree			or {[1] = LibInventory.types_base}

LibInventory.STATUS_INITIALIZING	= LibInventory.STATUS_INITIALIZING	or "INITIALIZING"
LibInventory.STATUS_QUERYING		= LibInventory.STATUS_QUERYING		or "QUERYING"
LibInventory.STATUS_SCANNING		= LibInventory.STATUS_SCANNING		or "SCANNING"
LibInventory.STATUS_READY			= LibInventory.STATUS_READY			or "READY"
LibInventory.status					= LibInventory.status				or LibInventory.STATUS_INITIALIZING

-- -----
-- locale stuff
-- -----
LibInventory.LOCALE_INVTYPE_OTHER = "Other"

local client_locale = GetLocale()
if client_locale == "deDE" then
	LibInventory.LOCALE_INVTYPE_OTHER = "Sonstiges"
end


-- -----
-- local helpers
-- -----
local CONST_ITEMS_PER_SCAN_SLOW = 50
local CONST_ITEMS_PER_SCAN_FAST = 1000
local CONST_ITEMS_MAX_ID = 91016 -- 5.0.4

local CONST_ITEMS_PER_SCAN = CONST_ITEMS_PER_SCAN_SLOW

local PATTERN_itemid = "item:(%-?%d+):%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+"
local PATTERN_itemlink = "|c%x+|H"..PATTERN_itemid.."|h%[.-%]|h|r"
local FORMAT_itemid = "item:%d:0:0:0:0:0:0:0:0:0"

local items = LibInventory.items
local items_valid_ids = LibInventory.items_valid_ids
local items_rescan = LibInventory.items_rescan

local function getbaseid(link)
	if link then
		if type(link) == "number" then return link end
		local baseid = strmatch(link, PATTERN_itemlink)
		if not baseid then
			baseid = strmatch(link, PATTERN_itemid)
		end
		if not baseid then
			baseid = link
		end
		return tonumber(baseid)
	end
	return
end


-- -----
-- public API
-- -----
LibInventory.GetItemBaseId = getbaseid

function LibInventory:GetItems(force_rescan)
	if force_rescan then
		self:BuildLists(true)
	end
	return self.items
end

function LibInventory:GetTypesTables()
	return self.types_tree, self.types_base, self.types_sub, self.types_slot
end

function LibInventory:GetCounts()
	return self.count_known, self.count_total, self.highest_id
end

function LibInventory:QueryItem(item)
	item = getbaseid(item)
	if not item then return end
	self:BuildLists(false, item)
end

function LibInventory:GetStatus()
	return self.status
end

function LibInventory:ActivateFastScan()
	CONST_ITEMS_PER_SCAN = CONST_ITEMS_PER_SCAN_FAST
end


-- -----
-- data storing
-- -----
function LibInventory:BuildLists(force_full_scan, scanid)
	if (self.status == self.STATUS_INITIALIZING) then
		local count_total = self.count_total
		local highest_id = self.highest_id
		local last_scanned = self.item_last_scanned or 0
		local max_scan = (not (force_full_scan or scanid)) and min(last_scanned + 2500, CONST_ITEMS_MAX_ID) or CONST_ITEMS_MAX_ID
		for i = last_scanned + 1, max_scan do
			if GetItemIcon(i) then
				items_valid_ids[i] = true
				count_total = count_total + 1
				highest_id = i
			end
		end
		self.item_last_scanned = max_scan
		self.count_total = count_total
		self.highest_id = highest_id
		if (self.item_last_scanned == CONST_ITEMS_MAX_ID) then
			self.item_last_scanned = nil
			self.status = self.STATUS_QUERYING
			self.callbacks:Fire("LibInventory_StatusChanged", self.status)
		end
		if (not (force_full_scan or scanid)) then return end
	end

	local newtype = false
	local scancount = 0
	local item_last_scanned_backup = self.item_last_scanned
	if force_full_scan == true then
		self.item_last_scanned = nil
		scanid = nil
	end
	local types_tree = self.types_tree
	local count_known = self.count_known
	local id, should_scan
	repeat
		if scanid then
			should_scan = true
			id = scanid
		else
			id = next(items_valid_ids, self.item_last_scanned)
			self.item_last_scanned = id
			if not id then -- full scan run complete
				collectgarbage("collect")
				if (self.status == self.STATUS_QUERYING) then
					self.status = self.STATUS_SCANNING
					self.callbacks:Fire("LibInventory_StatusChanged", self.status)
				elseif (self.status == self.STATUS_SCANNING) then
					self.status = self.STATUS_READY
					self.callbacks:Fire("LibInventory_StatusChanged", self.status)
					self.frame:UnregisterAllEvents()
					self.frame:Hide()
				end
				break
			end
			scancount = scancount + 1
		end
		if (items[id] == nil) then
--			local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture, itemSellPrice = GetItemInfo(id)
			local itemName, _, _, _, _, itemType, itemSubType, _, itemEquipLoc, _, _ = GetItemInfo(id)
			if (itemName and itemName ~= "") then -- cached item
				itemType, itemSubType, itemEquipLoc = tostring(itemType), tostring(itemSubType), _G[itemEquipLoc]
				if (not itemType or itemType == "") then
					itemType = self.LOCALE_INVTYPE_OTHER
				end
				if (not itemSubType or itemSubType == "") then
					itemSubType = self.LOCALE_INVTYPE_OTHER
				end
				if (not itemEquipLoc or itemEquipLoc == "") then
					itemEquipLoc = self.LOCALE_INVTYPE_OTHER
				end
				if (items[id] == nil) then
					items[id] = false
					count_known = count_known + 1
					if (count_known % 10000 == 0) then
						collectgarbage("collect")
					end
				end
				if (not types_tree[itemType]) then
					types_tree[itemType] = {}
					newtype = true
				end
				if (not types_tree[itemType][itemSubType]) then
					types_tree[itemType][itemSubType] = {}
					newtype = true
				end
				if (not types_tree[itemType][itemSubType][itemEquipLoc]) then
					types_tree[itemType][itemSubType][itemEquipLoc] = true
					newtype = true
				end
				items_valid_ids[id] = nil
			else --item wasn't cached, yet; GetItemInfo call caused a request
				if (items_valid_ids[id]) then
					items_valid_ids[id] = false
				else
					items_valid_ids[id] = nil
					items_rescan[id] = true
				end
			end
		end
	until not force_full_scan and (scanid or scancount >= CONST_ITEMS_PER_SCAN)
	self.count_known = count_known
	if force_full_scan == true then
		self.item_last_scanned = item_last_scanned_backup
	end
	if newtype then self:UpdateTypesTree() end
end

function LibInventory:UpdateTypesTree()
	local types_tree = self.types_tree
	local types_base = table_wipe(self.types_base)
	local types_sub = table_wipe(self.types_sub)
	local types_slot = table_wipe(self.types_slot)

	for basetype, subtypes in pairs(types_tree) do
		if basetype ~= 1 then
			table_insert(types_base, basetype)
			local subtypesort = subtypes[1] and table_wipe(subtypes[1]) or {}
			local hasinvtypes = false
			for subtype, invtypes in pairs(subtypes) do
				if subtype ~= 1 then
					table_insert(subtypesort, tostring(subtype))
					local invtypesort = invtypes[1] and table_wipe(invtypes[1]) or {}
					for invtype, _ in pairs(invtypes) do
						if invtype ~= 1 then
							table_insert(invtypesort, tostring(invtype))
						end
					end
					if #invtypesort > 1 then
						table_sort(invtypesort)
						if (not types_slot[basetype]) then
							types_slot[basetype] = {}
						end
						types_slot[basetype][subtype] = invtypesort
						hasinvtypes = true
					end
					invtypes[1] = invtypesort
				end
			end
			if #subtypesort > 1 or hasinvtypes then
				table_sort(subtypesort)
				types_sub[basetype] = subtypesort
			end
			subtypes[1] = subtypesort
		end
	end
	table_sort(self.types_base)
	types_tree[1] = types_base
	self.callbacks:Fire("LibInventory_TypesUpdated")
end


-- -----
-- OnEvent and OnUpdate handlers / event registering
-- -----
local eventmap = {
-- other
	["PLAYER_ENTERING_WORLD"] = "PLAYER_ENTERING_WORLD",
	["PLAYER_LEAVING_WORLD"] = "PLAYER_LEAVING_WORLD",
	["PLAYER_REGEN_ENABLED"] = "PLAYER_REGEN_ENABLED",
	["PLAYER_REGEN_DISABLED"] = "PLAYER_REGEN_DISABLED",
}

LibInventory.frame:SetScript("OnEvent", function(self, event, ...)
	local eventfunc = eventmap[event]
	if eventfunc then
		LibInventory[eventfunc](LibInventory, ...)
	end
end)

for event in pairs(eventmap) do
	LibInventory.frame:RegisterEvent(event)
end

local update_throttle = 0
LibInventory.frame:SetScript("OnUpdate", function(self, elapsed)
	update_throttle = update_throttle - elapsed
	if (update_throttle > 0) then return end
	update_throttle = 0.1
	LibInventory:BuildLists()
end)

LibInventory.frame:Hide()


-- -----
-- event functions
-- -----
function LibInventory:PLAYER_ENTERING_WORLD()
--~ 	self:BuildLists(true)
	if not InCombatLockdown() then
		LibInventory.frame:Show()
	end
end

function LibInventory:PLAYER_LEAVING_WORLD()
	LibInventory.frame:Hide()
end

function LibInventory:PLAYER_REGEN_ENABLED()
	LibInventory.frame:Show()
end

function LibInventory:PLAYER_REGEN_DISABLED()
	LibInventory.frame:Hide()
end
